\begin{code}
-- # import ExamplePrelude
\end{code}

 An example from M. Beeson \cite{beeson:otterLambda}. This example illustrates
how we can combine the interactive style of the logical framework, for instance
for the induction steps, with the first-order logic plugin.

 In this example we want to reason about existentially quantified propositions
so we add some new constants to the signature.

\begin{code}
Exists (A:Set) : (A -> Prop) -> Prop
existsI (A:Set) : (P:A->Prop) -> (x:A) -> P x -> Exists P
existsE (A:Set) : (P:A->Prop) -> Exists P ->
		    (C:Prop) -> ((x:A) -> P x ==> C) -> C
\end{code}

 We also need natural numbers. For this use the datatype extensions which
allows us to define recursive functions over the natural numbers. For
instance, we can write a recursive proof of the induction principle.

\begin{code}

data Nat : Set where
 zero : Nat
 succ : Nat -> Nat

indNat : (P:Nat -> Prop) -> P zero ->
	  ((n:Nat) -> P n ==> P (succ n)) ->
	  (x:Nat) -> P x
indNat P a g zero = a
indNat P a g (succ n) = impE (g n) (indNat P a g n)

\end{code}

 The goal of the example is to prove that in an integral ring, the only
nilpotent element is zero. We start by defining what it means to be an
integral ring.

\begin{code}

-- general definitions

isRing : (R:Set) -> (R->R->R) -> (R->R->R) ->
		      (R->R) -> R -> R -> Prop
isRing R (+) (*) minus Zero One =
 (x:R) -> (y:R) -> (z:R) ->
  (   (x + y) == (y + x)
   /\ (x + Zero) == x
   /\ (x + (minus x)) == Zero
   /\ (x + (y+z)) == ((x+y)+z)
   /\ (x * (y+z)) == ((x*y) + (x*z))
   /\ ((y+z)* x) == ((y*x) + (z*x))
   /\ (x * One) == x
   /\ (One * x) == x
   /\ (x * (y*z)) == ((x*y)*z)
  )

isIntegral : (R:Set) -> (R->R->R) -> R -> Prop
isIntegral R (*) Zero =
    (x:R) -> (y:R) -> x*y == Zero ==>
			x == Zero \/ y == Zero
\end{code}

In the following we work on a particular (but abstract) integral ring.

\begin{code}
-- the example

R : Set
(+) : R -> R -> R
(*) : R -> R -> R
minus : R -> R 
Zero : R
One : R

axR : isRing  R (+) (*) minus Zero One 
axI : isIntegral R (*) Zero

power : Nat -> R -> R
power zero x = One
power (succ n) x = (power n x) * x

isZero : R -> Prop
isZero x = x == Zero

isNilpotent : R -> Prop
isNilpotent x = Exists (\ n -> isZero (power n x))

\end{code}

This is all we need to start the proof. First we prove some lemmas.

\begin{code}

lemCancel : (x:R) -> (y:R) -> x + y == y ==> isZero x
lemCancel x y =
  impI (\h ->
	   let rem : isZero (x + (y + minus y)) 
	       rem = fol-plugin(h,axR)
	   in
	     fol-plugin(rem,axR)
       )

\end{code}

 The proof of $Zero * x \mathrel{=\mkern-4mu=} Zero$ is not trivial (but can
be done purely automatically if desired) so we give the main steps of one
possible proof explicitely.

\begin{code}

lemZero : (x:R) -> isZero (Zero * x) 
lemZero x =
  let rem1 : Zero + One == One
      rem1 = fol-plugin(axR)
      rem2 : (Zero + One) * x == Zero * x + One * x 
      rem2 = fol-plugin(axR)
      rem3 : Zero * x + One * x == One * x
      rem3 = fol-plugin(axR,rem1,rem2)
  in
    fol-plugin(rem3,lemCancel)

lemOneZero : (x:R) -> One == Zero ==> isZero x
lemOneZero x = fol-plugin(axR,lemZero)

\end{code}

 The main lemma is proved by induction explicitely at the framework level.

\begin{code}

prop : R -> Nat -> Prop
prop x n = isZero (power n x) ==> isZero x

lemMain : (x:R) -> (n:Nat) -> prop x n
lemMain x =
  let base : prop x zero
      base = fol-plugin(lemOneZero)
      step : (n:Nat) -> prop x n ==> prop x (succ n)
      step n = fol-plugin(axR,axI)
  in
    indNat (prop x) base step

thm : (x:R) -> isNilpotent x -> isZero x
thm x h = existsE (\n -> isZero (power n x)) h (isZero x) (lemMain x)

\end{code}




% vim: tw=78

