
module Plugins.Time where

import Prelude hiding (print, putStrLn, putStr)
import IO.Output

import Control.Monad.Trans
import GHC.Float

import Syntax.Position
import Syntax.Internal
import Syntax.Values
import Syntax.Pretty

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import TypeChecker

import Utils.Error
import Utils.Time

import Plugins

plugin :: SomePlugin
plugin = mkPlugin $ 
    Plugin  { plName		= "time"
            , plHelp            = help
	    , plInit		= initialize
	    , plFinal		= finalize
	    , plCheck		= check
	    , plInfer		= infer
	    , plOptions		= options
	    , plDefaultOptions	= defaultOptions
	    }

type State = Integer
type Options = ()

help = "Measures the time it takes to type check its argument."

initialize :: Options -> TC State
initialize _ = liftIO currentTimeMillis

finalize :: Options -> State -> TC ()
finalize _ _ = return ()

time :: State -> Doc -> TC a -> TC a
time t0 lbl m =
    do	t1 <- liftIO currentTimeMillis
	x  <- m
	t2 <- liftIO currentTimeMillis
	let show' x = text (formatRealFloat FFFixed (Just 2) $ fromIntegral x / 1000) <> text "s"
	    start = show' $ t1 - t0
	    end   = show' $ t2 - t0
	    delta = show' $ t2 - t1
	liftIO $ print $ hsep [ delta
			      , lbl
			      , parens $ start <+> text "to" <+> end
			      ]
	return x

check :: Options -> State -> Range -> [Term] -> Type -> TC (Value, State)
check _ s _ es t =
    case es of
	[e]	-> time s (pretty e) $ chk e
	[lbl,e] -> time s (pretty lbl) $ chk e
	_	-> failure $ customError "The time-plugin takes one or two arguments"
    where
	chk e = do v <- checkType e t
		   return (v,s)

infer :: Options -> State -> Range -> [Term] -> TC (Value, Type, State)
infer _ s _ es =
    case es of
	[e]	-> time s (pretty e) $ chk e
	[lbl,e] -> time s (pretty lbl) $ chk e
	_	-> failure $ customError "The time-plugin takes one or two arguments"
    where
	chk e = do (v,t) <- inferType e
		   return (v,t,s)

options :: PluginOptions Options
options = simpleOptions []

defaultOptions :: Options
defaultOptions = ()

